"""Plotting methods."""
from typing import Dict, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np

from eddington.exceptions import PlottingError
from eddington.fitting_data import FittingData
from eddington.print_util import to_precise_string


def plot_residuals(  # pylint: disable=invalid-name,too-many-arguments
    func,
    data: FittingData,
    a: np.ndarray,
    title_name,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    grid: bool = False,
    xmin: Optional[float] = None,
    xmax: Optional[float] = None,
):
    """
    Plot residuals plot.

    :param func: Fitting function.
    :type func: :class:`FittingFunction`
    :param data: Fitting data
    :type data: :class:`FittingData`
    :param a: The parameters result
    :type a: ``numpy.ndarray`` or ``list``
    :param title_name: Optional. Title for the figure.
    :type title_name: str
    :param xlabel: Optional. Label of the x axis
    :type xlabel: str
    :param ylabel: Optional. Label of the x axis
    :type ylabel: str
    :param grid: Add grid lines or not
    :type grid: bool
    :param xmin: Optional. minimum value for x in plot
    :type xmin: float
    :param xmax: Optional. maximum value for x in plot
    :type xmax: float
    :returns: ``matplotlib.pyplot.Figure``
    """
    fig = plot_data(
        data=data.residuals(func, a),
        title_name=title_name,
        xlabel=xlabel,
        ylabel=ylabel,
        grid=grid,
    )
    xmin, xmax = get_plot_borders(x=data.x, xmin=xmin, xmax=xmax)
    horizontal_line(fig=fig, xmin=xmin, xmax=xmax)
    return fig


def plot_fitting(  # pylint: disable=C0103,R0913,R0914
    func,
    data: FittingData,
    a: Union[np.ndarray, List[np.ndarray], Dict[str, np.ndarray]],
    title_name,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    grid: bool = False,
    legend: Optional[bool] = None,
    step: Optional[float] = None,
    xmin: Optional[float] = None,
    xmax: Optional[float] = None,
):
    """
    Plot fitting plot.

    :param func: Fitting function.
    :type func: :class:`FittingFunction`
    :param data: Fitting data
    :type data: :class:`FittingData`
    :param a: The parameters result
    :type a: ``numpy.ndarray``, a list of ``numpy.ndarray`` items or a dictionary from
     strings to ``numpy.ndarray``
    :param title_name: Optional. Title for the figure.
    :type title_name: str
    :param xlabel: Optional. Label of the x axis
    :type xlabel: str
    :param ylabel: Optional. Label of the x axis
    :type ylabel: str
    :param grid: Add grid lines or not
    :type grid: bool
    :param legend: Add legend or not
    :type legend: bool
    :param step: Optional. Steps between samples for the fitting graph
    :type step: float
    :param xmin: Optional. minimum value for x in plot
    :type xmin: float
    :param xmax: Optional. maximum value for x in plot
    :type xmax: float
    :returns: ``matplotlib.pyplot.Figure``
    """
    fig = plot_data(
        data=data, title_name=title_name, xlabel=xlabel, ylabel=ylabel, grid=grid
    )
    xmin, xmax = get_plot_borders(x=data.x, xmin=xmin, xmax=xmax)
    if step is None:
        step = (xmax - xmin) / 1000.0
    x = np.arange(xmin, xmax, step=step)  # pylint: disable=invalid-name
    a_dict = __get_a_dict(a)
    for label, a_value in a_dict.items():
        plot(fig=fig, x=x, y=func(a_value, x), label=label)
    if __get_legend(legend, a_dict):
        plt.legend()
    return fig


def plot_data(
    data: FittingData,
    title_name,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    grid: bool = False,
):
    """
    Plot fitting data.

    :param data: Fitting data
    :type data: :class:`FittingData`
    :param title_name: Optional. Title for the figure.
    :type title_name: str
    :param xlabel: Optional. Label of the x axis
    :type xlabel: str
    :param ylabel: Optional. Label of the x axis
    :type ylabel: str
    :param grid: Add grid lines or not
    :type grid: bool
    :returns: ``matplotlib.pyplot.Figure``
    """
    fig = get_figure(title_name=title_name, xlabel=xlabel, ylabel=ylabel, grid=grid)
    errorbar(fig=fig, x=data.x, y=data.y, xerr=data.xerr, yerr=data.yerr)
    return fig


def get_figure(
    title_name,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    grid: bool = False,
):
    """
    Gets a figure from matplotlib.

    :param title_name: Optional. Title for the figure.
    :param xlabel: Optional. Label of the x axis
    :param ylabel: Optional. Label of the x axis
    :param grid: Add grid lines or not
    :return: Figure instance
    """
    fig = plt.figure()
    title(fig=fig, title_name=title_name)
    label_axes(fig=fig, xlabel=xlabel, ylabel=ylabel)
    add_grid(fig=fig, is_grid=grid)
    return fig


def title(fig, title_name):
    """
    Add/remove title to figure.

    :param title_name: Optional. If None, don't add title. otherwise, add given title
    :param fig: Plot figure.
    """
    if title_name is not None:
        plt.title(title_name, figure=fig)


def label_axes(fig, xlabel, ylabel):
    """
    Add/remove labels to figure.

    :param fig: Plot figure.
    :param xlabel: Optional. If None, don't add label. otherwise, add given label
    :param ylabel: Optional. If None, don't add label. otherwise, add given label
    """
    if xlabel is not None:
        plt.xlabel(xlabel, figure=fig)
    if ylabel is not None:
        plt.ylabel(ylabel, figure=fig)


def add_grid(fig, is_grid):
    """
    Add/remove grid to figure.

    :param fig: Plot figure
    :param is_grid: Boolean. add or remote grid to plot
    """
    if is_grid:
        plt.grid(True, figure=fig)


def plot(x, y, fig, label=None):  # pylint: disable=C0103
    """
    Plot y as a function of x.

    :param x: X values
    :param y: Y values
    :param fig: Plot figure
    :param label: Optional. Label for the plot that would be added to the legend
    """
    plt.plot(x, y, figure=fig, label=label)


def horizontal_line(  # pylint: disable=C0103
    fig: plt.Figure, xmin: float, xmax: float, y=0
):
    """
    Add horizontal line to figure.

    :param xmin: Minimum x value of line
    :param xmax: Maximum x value of line
    :param y: The y value of the line
    :param fig: Plot figure
    """
    plt.hlines(y, xmin=xmin, xmax=xmax, linestyles="dashed", figure=fig)


def errorbar(fig, x, y, xerr, yerr):  # pylint: disable=C0103
    """
    Plot error bar to figure.

    :param x: X values
    :param y: Y values
    :param xerr: Errors of x
    :param yerr: Errors of y
    :param fig: Plot figure
    """
    plt.errorbar(
        x=x,
        y=y,
        xerr=xerr,
        yerr=yerr,
        markersize=1,
        marker="o",
        linestyle="None",
        figure=fig,
    )


def get_plot_borders(  # pylint: disable=invalid-name
    x: np.ndarray, xmin: Optional[float] = None, xmax: Optional[float] = None
) -> Tuple[float, float]:
    """
    Get borders for a plot based on its x values.

    :param x: x values list or array.
    :type x: ``numpy.ndarray`` or ``list``
    :param xmin: Minimum x value of line
    :type xmin: float
    :param xmax: Maximum x value of line
    :type xmax: float
    :return: tuple. minimum and maximum values for the plot.
    """
    data_xmin = np.min(x)
    data_xmax = np.max(x)
    gap = (data_xmax - data_xmin) * 0.1
    if xmin is None:
        xmin = data_xmin - gap
    if xmax is None:
        xmax = data_xmax + gap
    return xmin, xmax


def show_or_export(fig: plt.Figure, output_path=None):
    """
    Show plot or export it to a file.

    :param fig: a plot figure
    :param output_path: Path or None. if None, show plot. otherwise, save to path.
    """
    if output_path is None:
        plt.show()
        return
    fig.savefig(output_path)


def __get_a_dict(a):  # pylint: disable=invalid-name
    if isinstance(a, dict):
        return a
    if isinstance(a, list):
        return {__build_repr_string(a_value): a_value for a_value in a}
    if isinstance(a, np.ndarray):
        return {__build_repr_string(a): a}
    raise PlottingError(
        f"{a} has unmatching type. Can except only numpy arrays, "
        "lists of numpy arrays and dictionaries."
    )


def __build_repr_string(a):  # pylint: disable=invalid-name
    arguments_values = [f"a[{i}]={to_precise_string(val)}" for i, val in enumerate(a)]
    return f"[{', '.join(arguments_values)}]"


def __get_legend(legend, a):  # pylint: disable=invalid-name
    if legend is not None:
        return legend
    if len(a) >= 2:
        return True
    return False



"""List of constants used by Eddington."""
DEFAULT_MIN_COEFF = -10.0
DEFAULT_MAX_COEFF = 10.0
DEFAULT_MEASUREMENTS = 20
DEFAULT_XMIN = -20.0
DEFAULT_XMAX = 20.0
DEFAULT_XSIGMA = 0.5
DEFAULT_YSIGMA = 0.5

DEFAULT_PRECISION = 3



"""Registry containing predefined fitting functions and fitting functions generators."""
from typing import List

from eddington.exceptions import FittingFunctionLoadError, FittingFunctionSaveError


class FittingFunctionsRegistry:  # noqa: D415,D213,D205
    """A singleton class containing all saved :class:`FittingFunction` instances."""

    __name_to_func_dict = dict()  # type: ignore

    @classmethod
    def add(cls, func):  # type: ignore
        """
        Add a fitting function.

        :param func: fitting function to add to registry
        :type func: :class:`FittingFunction`
        """
        if func.name in cls.__name_to_func_dict:
            raise FittingFunctionSaveError(
                f'Cannot save "{func.name}" to registry'
                " since there is another fitting function with this name"
            )
        cls.__name_to_func_dict[func.name] = func

    @classmethod
    def remove(cls, func_name: str) -> None:
        """
        Remove a fitting function.

        :param func_name: Name of the function to remove.
        :type func_name: str
        """
        del cls.__name_to_func_dict[func_name]

    @classmethod
    def clear(cls) -> None:
        """Clear all fitting functions and generators."""
        cls.__name_to_func_dict.clear()

    @classmethod
    def all(cls):  # type: ignore
        """
        Get all fitting functions and generators.

        :returns: list of :class:`FittingFunction`
        """
        return list(cls.__name_to_func_dict.values())

    @classmethod
    def names(cls) -> List[str]:
        """Names of all fitting functions and generators."""
        return list(cls.__name_to_func_dict.keys())

    @classmethod
    def load(cls, func_name):  # type: ignore
        """
        Get a fitting function and generators by name.

        :param func_name: Name of the function to load.
        :type func_name: str
        :returns: :class:`FittingFunction`
        """
        if not cls.exists(func_name):
            raise FittingFunctionLoadError(f"No fitting function named {func_name}")
        return cls.__name_to_func_dict[func_name]

    @classmethod
    def exists(cls, func_name: str) -> bool:
        """
        Checks whether a fitting function exist.

        :param func_name: Name of the function to load.
        :type func_name: str
        :returns: bool
        """
        return func_name in cls.__name_to_func_dict



"""Implementation of the fitting algorithm."""
from typing import Any, Dict, Optional

import numpy as np
from scipy.odr import ODR, Model, RealData

from eddington.fitting_data import FittingData
from eddington.fitting_function_class import FittingFunction
from eddington.fitting_result import FittingResult


def fit(  # pylint: disable=invalid-name
    data: FittingData,
    func: FittingFunction,
    a0: np.ndarray = None,
    use_x_derivative: bool = True,
    use_a_derivative: bool = True,
) -> FittingResult:
    """
    Implementation of the fitting algorithm.

    This functions wraps *scipy*'s
    `ODR <https://docs.scipy.org/doc/scipy/reference/odr.html>`_ algorithm.

    :param data: Fitting data to optimize
    :type data: :class:`FittingData`
    :param func: a function to fit the data according to.
    :type func: :class:`FittingFunction`
    :param a0: initial guess for the parameters
    :type a0: ``np.ndarray``
    :param use_x_derivative: indicates whether to use x derivative or not.
    :type use_x_derivative: ``bool``
    :param use_a_derivative: indicates whether to use a derivative or not.
    :type use_a_derivative: ``bool``
    :returns: :class:`FittingResult`
    """
    model = Model(
        **__get_odr_model_kwargs(
            func,
            use_x_derivative=use_x_derivative,
            use_a_derivative=use_a_derivative,
        )
    )
    a0 = __get_a0(n=func.active_parameters, a0=a0)
    real_data = RealData(x=data.x, y=data.y, sx=data.xerr, sy=data.yerr)
    odr = ODR(data=real_data, model=model, beta0=a0)
    output = odr.run()
    a = output.beta  # pylint: disable=invalid-name
    chi2 = output.sum_square  # pylint: disable=no-member
    degrees_of_freedom = len(data.x) - func.active_parameters
    return FittingResult(
        a0=a0,
        a=a,
        aerr=output.sd_beta,
        acov=output.cov_beta,
        degrees_of_freedom=degrees_of_freedom,
        chi2=chi2,
    )


def __get_odr_model_kwargs(
    func: FittingFunction,
    use_x_derivative: bool = True,
    use_a_derivative: bool = True,
) -> Dict[str, Any]:
    kwargs: Dict[str, Any] = dict(fcn=func)
    if use_a_derivative and func.a_derivative is not None:
        kwargs["fjacb"] = func.a_derivative
    if use_x_derivative and func.x_derivative is not None:
        kwargs["fjacd"] = func.x_derivative
    return kwargs


def __get_a0(  # pylint: disable=invalid-name
    n: int, a0: Optional[np.ndarray] = None
) -> np.ndarray:
    """
    Returns initial parameters for fitting algorithm.

    :param n: Number of parameters
    :param a0: Initial parameters value. Optional
    :return: nd.array
    """
    if a0 is not None:
        return a0
    return np.full(shape=n, fill_value=1.0)



#!/usr/bin/env python
import os

from setuptools import setup

READTHEDOCS = "READTHEDOCS"

version = "0.0.17"
on_rtd = os.environ.get(READTHEDOCS) == "True"
install_requires = ["click >= 7.1.2", "ptable >= 0.7.2"]
if not on_rtd:
    install_requires.extend(
        [
            "scipy >= 1.5.2",
            "numpy >= 1.19.1",
            "openpyxl >= 3.0.5",
            "matplotlib >= 3.3.1",
        ]
    )
setup(version=version, install_requires=install_requires)



"""Fitting result class that will be returned by the fitting algorithm."""
import json
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Optional, Union

import numpy as np
import scipy.stats as stats

from eddington.consts import DEFAULT_PRECISION
from eddington.print_util import to_precise_string


@dataclass(repr=False)  # pylint: disable=too-many-instance-attributes
class FittingResult:
    """
    Dataclass that contains the relevant parameters returned by a fitting algorithm.

    :param a0: The initial guess for the fitting function parameters.
    :param a: The result for the fitting parameters.
    :param aerr: Estimated errors of a.
    :param arerr: Estimated relative errors of a (equivalent to aerr/a).
    :param acov: Covariance matrix of a.
    :param degrees_of_freedom: How many degrees of freedom of the fittings.
    :param chi2: Optimization evaluation for the fit.
    :param chi2_reduced: Reduced chi2.
    :param p_probability: P-probability (p-value) of the fitting, evaluated from
     chi2_reduced.
    """

    a0: Union[List[float], np.ndarray]  # pylint: disable=invalid-name
    a: Union[List[float], np.ndarray]  # pylint: disable=invalid-name
    aerr: Union[List[float], np.ndarray]
    arerr: Union[List[float], np.ndarray] = field(
        init=False,
    )
    acov: Union[List[List[float]], np.ndarray]
    degrees_of_freedom: int
    chi2: float
    chi2_reduced: float = field(init=False)
    p_probability: float = field(init=False)
    precision: int = field(default=DEFAULT_PRECISION)
    __pretty_string: Optional[str] = field(default=None, init=False)

    def __post_init__(self) -> None:
        """Post init methods."""
        self.aerr = np.array(self.aerr)
        self.acov = np.array(self.acov)
        self.a0 = np.array(self.a0)
        self.a = np.array(self.a)
        with np.errstate(divide="ignore"):
            self.arerr = np.abs(self.aerr / self.a) * 100
        self.chi2_reduced = self.chi2 / self.degrees_of_freedom
        self.p_probability = stats.chi2.sf(self.chi2, self.degrees_of_freedom)

    def save_txt(self, file_path: Union[str, Path]) -> None:
        """
        Write the result to a text file.

        :param file_path: Path to write the result in. if None, prints to
         console.
        :type file_path: ``str`` or ``Path``
        """
        with open(file_path, mode="w") as output_file:
            output_file.write(self.pretty_string)

    def save_json(self, file_path: Union[str, Path]) -> None:
        """
        Write the result to a json file.

        :param file_path: Path to write the result in. if None, prints to
         console.
        :type file_path: ``str`` or ``Path``
        """
        with open(file_path, mode="w") as output_file:
            json.dump(
                dict(
                    a0=self.a0.tolist(),  # type: ignore
                    a=self.a.tolist(),  # type: ignore
                    aerr=self.aerr.tolist(),  # type: ignore
                    arerr=self.arerr.tolist(),  # type: ignore
                    acov=self.acov.tolist(),  # type: ignore
                    degrees_of_freedom=self.degrees_of_freedom,
                    chi2=self.chi2,
                    chi2_reduced=self.chi2_reduced,
                    p_probability=self.p_probability,
                ),
                output_file,
                indent=1,
            )

    @property
    def pretty_string(self) -> str:
        """Pretty representation string."""
        if self.__pretty_string is None:
            self.__pretty_string = self.__build_pretty_string()
        return self.__pretty_string

    def __repr__(self) -> str:
        """Representation string."""
        return self.pretty_string

    def __build_pretty_string(self) -> str:
        old_precision = np.get_printoptions()["precision"]
        np.set_printoptions(precision=self.precision)
        a_value_string = "\n".join(
            [
                self.__a_value_string(i, a, aerr, arerr)
                for i, (a, aerr, arerr) in enumerate(zip(self.a, self.aerr, self.arerr))
            ]
        )
        repr_string = f"""Results:
========

Initial parameters' values:
\t{" ".join(str(i) for i in self.a0)}
Fitted parameters' values:
{a_value_string}
Fitted parameters covariance:
{self.acov}
Chi squared: {to_precise_string(self.chi2, self.precision)}
Degrees of freedom: {self.degrees_of_freedom}
Chi squared reduced: {to_precise_string(self.chi2_reduced, self.precision)}
P-probability: {to_precise_string(self.p_probability, self.precision)}
"""
        np.set_printoptions(precision=old_precision)
        return repr_string

    def __a_value_string(  # pylint: disable=invalid-name
        self, i: int, a: float, aerr: float, arerr: float
    ) -> str:
        a_string = to_precise_string(a, self.precision)
        aerr_string = to_precise_string(aerr, self.precision)
        arerr_string = to_precise_string(arerr, self.precision)
        return f"\ta[{i}] = {a_string} \u00B1 {aerr_string} ({arerr_string}% error)"



"""CLI for Eddington."""
import re
from pathlib import Path
from typing import Any, Dict, Optional, Union

import click
import numpy as np
from prettytable import PrettyTable

from eddington import __version__
from eddington.fitting import fit
from eddington.fitting_data import FittingData
from eddington.fitting_functions_list import linear, polynomial
from eddington.fitting_functions_registry import FittingFunctionsRegistry
from eddington.plot import plot_data, plot_fitting, plot_residuals, show_or_export

# pylint: disable=too-many-arguments,invalid-name,too-many-locals


@click.group("eddington")
@click.version_option(version=__version__)
def eddington_cli():
    """Command line for Eddington."""


@eddington_cli.command("list")
@click.option(
    "-r",
    "--regex",
    type=str,
    default=None,
    help="Filter functions by a regular expression",
)
def eddington_list(regex: Optional[str]):
    """Prints all fitting functions in a pretty table."""
    table = PrettyTable(field_names=["Function", "Syntax"])
    for func in FittingFunctionsRegistry.all():
        if regex is None or re.search(regex, func.name):
            table.add_row([func.name, func.syntax])
    click.echo(table)


@eddington_cli.command("fit")
@click.pass_context
@click.argument("fitting_function_name", type=str, default="")
@click.option(
    "-p",
    "--polynomial",
    "polynomial_degree",
    type=int,
    help="Fitting data to polynomial of nth degree.",
)
@click.option(
    "-d",
    "--data-file",
    required=True,
    type=click.Path(exists=True, dir_okay=False, file_okay=True),
    help="Data file to read from.",
)
@click.option("-s", "--sheet", type=str, help="Sheet name for excel files.")
@click.option(
    "--a0",
    type=str,
    help=(
        "Initial guess for the fitting algorithm. "
        "Should be given as floating point numbers separated by commas"
    ),
)
@click.option("--x-column", type=str, help="Column to read x values from.")
@click.option("--xerr-column", type=str, help="Column to read x error values from.")
@click.option("--y-column", type=str, help="Column to read y values from.")
@click.option("--yerr-column", type=str, help="Column to read y error values from.")
@click.option("--x-label", type=str, help="Label for the x axis.")
@click.option("--y-label", type=str, help="Label for the y axis.")
@click.option("--grid/--no-grid", default=False, help="Add grid lines to plots.")
@click.option(
    "--plot-fitting/--no-plot-fitting",
    "should_plot_fitting",
    default=True,
    help="Should plot fitting.",
)
@click.option(
    "--plot-residuals/--no-plot-residuals",
    "should_plot_residuals",
    default=True,
    help="Should plot residuals.",
)
@click.option(
    "--plot-data/--no-plot-data",
    "should_plot_data",
    default=False,
    help="Should plot data.",
)
@click.option(
    "--legend/--no-legend",
    default=None,
    help="Should add legend to fitting plot.",
)
@click.option(
    "-o",
    "--output-dir",
    type=click.Path(dir_okay=True, file_okay=False),
    help="Output directory to save plots in.",
)
@click.option(
    "--json",
    is_flag=True,
    default=False,
    help="Save result as json instead of text.",
)
def eddington_fit(
    ctx: click.Context,
    fitting_function_name: Optional[str],
    polynomial_degree: Optional[int],
    data_file: str,
    sheet: Optional[str],
    a0: Optional[str],
    x_column: Optional[str],
    xerr_column: Optional[str],
    y_column: Optional[str],
    yerr_column: Optional[str],
    x_label: Optional[str],
    y_label: Optional[str],
    grid,
    should_plot_fitting: bool,
    should_plot_residuals: bool,
    should_plot_data: bool,
    legend: Optional[bool],
    output_dir: Union[Path, str],
    json: bool,
):
    """Fitting data file according to a fitting function."""
    # fmt: off
    data = __load_data_file(
        ctx, Path(data_file), sheet,
        x_column=x_column, xerr_column=xerr_column,
        y_column=y_column, yerr_column=yerr_column,
    )
    # fmt: on
    func = __load_fitting_functions(
        ctx=ctx, func_name=fitting_function_name, polynomial_degree=polynomial_degree
    )
    result = fit(data, func, a0=__calc_a0(a0))
    click.echo(result.pretty_string)
    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)
        if json:
            result.save_json(output_dir / f"{func.name}_result.json")
        else:
            result.save_txt(output_dir / f"{func.name}_result.txt")
    if x_label is None:
        x_label = data.x_column
    if y_label is None:
        y_label = data.y_column
    plot_kwargs: Dict[str, Any] = dict(xlabel=x_label, ylabel=y_label, grid=grid)
    if should_plot_data:
        show_or_export(
            plot_data(data=data, title_name=f"{func.title_name} - Data", **plot_kwargs),
            output_path=__optional_path(output_dir, f"{func.name}_data.png"),
        )
    if should_plot_fitting:
        show_or_export(
            plot_fitting(
                func=func,
                data=data,
                a=result.a,
                title_name=f"{func.title_name}",
                legend=legend,
                **plot_kwargs,
            ),
            output_path=__optional_path(output_dir, f"{func.name}.png"),
        )
    if should_plot_residuals:
        show_or_export(
            plot_residuals(
                func=func,
                data=data,
                a=result.a,
                title_name=f"{func.title_name} - Residuals",
                **plot_kwargs,
            ),
            output_path=__optional_path(output_dir, f"{func.name}_residuals.png"),
        )


def __calc_a0(a0):
    if a0 is None:
        return None
    return np.array(list(map(float, re.split(",[ \t]*", a0))))


def __load_data_file(
    ctx: click.Context, data_file: Path, sheet: Optional[str], **kwargs
):
    suffix = data_file.suffix
    if suffix == ".csv":
        return FittingData.read_from_csv(filepath=data_file, **kwargs)
    if suffix == ".json":
        return FittingData.read_from_json(filepath=data_file, **kwargs)
    if suffix != ".xlsx":
        click.echo(f'Cannot read data with "{suffix}" suffix')
        ctx.exit(1)
    if sheet is None:
        click.echo("Sheet name has not been specified!")
        ctx.exit(1)
    return FittingData.read_from_excel(filepath=data_file, sheet=sheet, **kwargs)


def __load_fitting_functions(
    ctx: click.Context, func_name: Optional[str], polynomial_degree: Optional[int]
):
    if func_name == "":
        if polynomial_degree is not None:
            return polynomial(polynomial_degree)
        return linear
    if polynomial_degree is not None:
        click.echo("Cannot accept both polynomial and fitting function")
        ctx.exit(1)
    return FittingFunctionsRegistry.load(func_name)


def __optional_path(directory: Optional[Path], file_name: str):
    if directory is None:
        return None
    return directory / file_name



"""Functions for data randomization."""
import numpy as np


def random_array(min_val: float, max_val: float, size: int) -> np.ndarray:
    """
    Creates a random array.

    :param min_val: Minimum value for array item.
    :param max_val: Maximum value for array item.
    :param size: Size of the array.
    :return: array
    """
    return np.random.uniform(min_val, max_val, size=size)


def random_sigma(average_sigma: float, size: int) -> np.ndarray:
    """
    Creates random standard deviation (aka. sigma).

    :param average_sigma: average value for sigma.
    :param size: Size of the returned array.
    :return: array of sigma values.
    """
    return np.random.exponential(average_sigma, size=size)


def random_error(scales: np.ndarray) -> np.ndarray:
    """
    Generates random errors based on an array of scales.

    :param scales: array of standard deviations.
    :return: errors array.
    """
    return np.random.normal(scale=scales)
